# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# NOTE! THIS FILE IS AUTOMATICALLY GENERATED AND WILL BE
# OVERWRITTEN WHEN PREPARING PACKAGES.
#
# IF YOU WANT TO MODIFY IT, YOU SHOULD MODIFY THE TEMPLATE
# `get_provider_info_TEMPLATE.py.jinja2` IN the `provider_packages` DIRECTORY

def get_provider_info():
    return {
        'package-name': 'apache-airflow-providers-google',
        'name': 'Google',
        'description': (
            'Google services including:\n\n  - `Google Ads <https://ads.google.com/>`__\n  - `Google Cloud'
            ' (GCP) <https://cloud.google.com/>`__\n  - `Google Firebase <https://firebase.google.com/>`__\n '
            ' - `Google LevelDB <https://github.com/google/leveldb/>`__\n  - `Google Marketing Platform'
            ' <https://marketingplatform.google.com/>`__\n  - `Google Workspace'
            ' <https://workspace.google.com/>`__ (formerly Google Suite)\n'
        ),
        'versions': [
            '8.3.0',
            '8.2.0',
            '8.1.0',
            '8.0.0',
            '7.0.0',
            '6.8.0',
            '6.7.0',
            '6.6.0',
            '6.5.0',
            '6.4.0',
            '6.3.0',
            '6.2.0',
            '6.1.0',
            '6.0.0',
            '5.1.0',
            '5.0.0',
            '4.0.0',
            '3.0.0',
            '2.2.0',
            '2.1.0',
            '2.0.0',
            '1.0.0',
        ],
        'dependencies': [
            'apache-airflow>=2.2.0',
            'apache-airflow-providers-common-sql>=1.1.0',
            'PyOpenSSL',
            'google-ads>=15.1.1',
            'google-api-core>=2.7.0,<3.0.0',
            'google-api-python-client>=1.6.0,<2.0.0',
            'google-auth>=1.0.0',
            'google-auth-httplib2>=0.0.1',
            'google-cloud-aiplatform>=1.7.1,<2.0.0',
            'google-cloud-automl>=2.1.0',
            'google-cloud-bigquery-datatransfer>=3.0.0',
            'google-cloud-bigtable>=1.0.0,<2.0.0',
            'google-cloud-build>=3.0.0',
            'google-cloud-container>=2.2.0,<3.0.0',
            'google-cloud-dataform>=0.2.0',
            'google-cloud-datacatalog>=3.0.0',
            'google-cloud-dataplex>=0.1.0',
            'google-cloud-dataproc>=3.1.0',
            'google-cloud-dataproc-metastore>=1.2.0,<2.0.0',
            'google-cloud-dlp>=0.11.0,<2.0.0',
            'google-cloud-kms>=2.0.0',
            'google-cloud-language>=1.1.1,<2.0.0',
            'google-cloud-logging>=2.1.1',
            'google-cloud-memcache>=0.2.0',
            'google-cloud-monitoring>=2.0.0',
            'google-cloud-os-login>=2.0.0',
            'google-cloud-orchestration-airflow>=1.0.0,<2.0.0',
            'google-cloud-pubsub>=2.0.0',
            'google-cloud-redis>=2.0.0',
            'google-cloud-secret-manager>=0.2.0,<2.0.0',
            'google-cloud-spanner>=1.10.0,<2.0.0',
            'google-cloud-speech>=0.36.3,<2.0.0',
            'google-cloud-storage>=1.30,<2.0.0',
            'google-cloud-tasks>=2.0.0',
            'google-cloud-texttospeech>=0.4.0,<2.0.0',
            'google-cloud-translate>=1.5.0,<2.0.0',
            'google-cloud-videointelligence>=1.7.0,<2.0.0',
            'google-cloud-vision>=0.35.2,<2.0.0',
            'google-cloud-workflows>=0.1.0,<2.0.0',
            'grpcio-gcp>=0.2.2',
            'httpx',
            'json-merge-patch>=0.2',
            'looker-sdk>=22.2.0',
            'pandas-gbq',
            'pandas>=0.17.1',
            'sqlalchemy-bigquery>=1.2.1',
            'proto-plus>=1.19.6',
        ],
        'integrations': [
            {
                'integration-name': 'Google Analytics360',
                'external-doc-url': 'https://analytics.google.com/',
                'logo': '/integration-logos/gcp/Google-Analytics.png',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/marketing_platform/analytics.rst'
                ],
                'tags': ['gmp'],
            },
            {
                'integration-name': 'Google Ads',
                'external-doc-url': 'https://ads.google.com/',
                'logo': '/integration-logos/gcp/Google-Ads.png',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/ads.rst'],
                'tags': ['gmp'],
            },
            {
                'integration-name': 'Google AutoML',
                'external-doc-url': 'https://cloud.google.com/automl/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/automl.rst'],
                'logo': '/integration-logos/gcp/Cloud-AutoML.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google BigQuery Data Transfer Service',
                'external-doc-url': 'https://cloud.google.com/bigquery/transfer/',
                'logo': '/integration-logos/gcp/BigQuery.png',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/bigquery_dts.rst'],
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google BigQuery',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/bigquery.rst'],
                'external-doc-url': 'https://cloud.google.com/bigquery/',
                'logo': '/integration-logos/gcp/BigQuery.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Bigtable',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/bigtable.rst'],
                'external-doc-url': 'https://cloud.google.com/bigtable/',
                'logo': '/integration-logos/gcp/Cloud-Bigtable.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Build',
                'external-doc-url': 'https://cloud.google.com/build/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/cloud_build.rst'],
                'logo': '/integration-logos/gcp/Cloud-Build.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Composer',
                'external-doc-url': 'https://cloud.google.com/composer',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/cloud_composer.rst'],
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Cloud Dataform',
                'external-doc-url': 'https://cloud.google.com/dataform/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/dataform.rst'],
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Cloud Data Loss Prevention (DLP)',
                'external-doc-url': 'https://cloud.google.com/dlp/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/data_loss_prevention.rst'
                ],
                'logo': '/integration-logos/gcp/google-data-loss-prevention.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Firestore',
                'external-doc-url': 'https://firebase.google.com/docs/firestore',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/firebase/firestore.rst'],
                'logo': '/integration-logos/gcp/Google-Firestore.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Functions',
                'external-doc-url': 'https://cloud.google.com/functions/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/functions.rst'],
                'logo': '/integration-logos/gcp/Cloud-Functions.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Key Management Service (KMS)',
                'external-doc-url': 'https://cloud.google.com/security-key-management/',
                'logo': '/integration-logos/gcp/Key-Management-Service.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Life Sciences',
                'external-doc-url': 'https://cloud.google.com/life-sciences/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/life_sciences.rst'],
                'logo': '/integration-logos/gcp/Google-Cloud-Life-Sciences.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Memorystore',
                'external-doc-url': 'https://cloud.google.com/memorystore/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/cloud_memorystore.rst',
                    '/docs/apache-airflow-providers-google/operators/cloud/cloud_memorystore_memcached.rst',
                ],
                'logo': '/integration-logos/gcp/Cloud-Memorystore.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud OS Login',
                'external-doc-url': 'https://cloud.google.com/compute/docs/oslogin/',
                'logo': '/integration-logos/gcp/Google-Cloud-Generic.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Pub/Sub',
                'external-doc-url': 'https://cloud.google.com/pubsub/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/pubsub.rst'],
                'logo': '/integration-logos/gcp/Cloud-PubSub.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Secret Manager',
                'external-doc-url': 'https://cloud.google.com/secret-manager/',
                'logo': '/integration-logos/gcp/Google-Cloud-Secret-Manager.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Spanner',
                'external-doc-url': 'https://cloud.google.com/spanner/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/spanner.rst'],
                'logo': '/integration-logos/gcp/Cloud-Spanner.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Speech-to-Text',
                'external-doc-url': 'https://cloud.google.com/speech-to-text/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/speech_to_text.rst',
                    '/docs/apache-airflow-providers-google/operators/cloud/translate_speech.rst',
                ],
                'logo': '/integration-logos/gcp/Cloud-Speech-to-Text.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud SQL',
                'external-doc-url': 'https://cloud.google.com/sql/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/cloud_sql.rst'],
                'logo': '/integration-logos/gcp/Cloud-SQL.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Stackdriver',
                'external-doc-url': 'https://cloud.google.com/products/operations/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/stackdriver.rst'],
                'logo': '/integration-logos/gcp/Google-Cloud-Stackdriver.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Storage (GCS)',
                'external-doc-url': 'https://cloud.google.com/storage/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/gcs.rst'],
                'logo': '/integration-logos/gcp/Cloud-Storage.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Tasks',
                'external-doc-url': 'https://cloud.google.com/tasks/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/tasks.rst'],
                'logo': '/integration-logos/gcp/Cloud-Tasks.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Text-to-Speech',
                'external-doc-url': 'https://cloud.google.com/text-to-speech/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/text_to_speech.rst'],
                'logo': '/integration-logos/gcp/Cloud-Text-to-Speech.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Translation',
                'external-doc-url': 'https://cloud.google.com/translate/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/translate.rst',
                    '/docs/apache-airflow-providers-google/operators/cloud/translate_speech.rst',
                ],
                'logo': '/integration-logos/gcp/Cloud-Translation-API.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Video Intelligence',
                'external-doc-url': 'https://cloud.google.com/video-intelligence/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/video_intelligence.rst'
                ],
                'logo': '/integration-logos/gcp/Cloud-Video-Intelligence-API.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Vision',
                'external-doc-url': 'https://cloud.google.com/vision/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/vision.rst'],
                'logo': '/integration-logos/gcp/Cloud-Vision-API.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Compute Engine',
                'external-doc-url': 'https://cloud.google.com/compute/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/compute.rst',
                    '/docs/apache-airflow-providers-google/operators/cloud/compute_ssh.rst',
                ],
                'logo': '/integration-logos/gcp/Compute-Engine.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Data Proc',
                'external-doc-url': 'https://cloud.google.com/dataproc/',
                'logo': '/integration-logos/gcp/Google-Data-Proc.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Data Catalog',
                'external-doc-url': 'https://cloud.google.com/data-catalog/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/datacatalog.rst'],
                'logo': '/integration-logos/gcp/Google-Data-Catalog.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Dataflow',
                'external-doc-url': 'https://cloud.google.com/dataflow/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/dataflow.rst'],
                'logo': '/integration-logos/gcp/Cloud-Dataflow.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Data Fusion',
                'external-doc-url': 'https://cloud.google.com/data-fusion/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/datafusion.rst'],
                'logo': '/integration-logos/gcp/Google-Data-Fusion.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Dataplex',
                'external-doc-url': 'https://cloud.google.com/dataplex/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/dataplex.rst'],
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Dataprep',
                'external-doc-url': 'https://cloud.google.com/dataprep/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/dataprep.rst'],
                'logo': '/integration-logos/gcp/Google-Dataprep.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Dataproc Metastore',
                'external-doc-url': 'https://cloud.google.com/dataproc-metastore/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/dataproc_metastore.rst'
                ],
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Dataproc',
                'external-doc-url': 'https://cloud.google.com/dataproc/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/dataproc.rst'],
                'logo': '/integration-logos/gcp/Cloud-Dataproc.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Datastore',
                'external-doc-url': 'https://cloud.google.com/datastore/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/datastore.rst'],
                'logo': '/integration-logos/gcp/Cloud-Datastore.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Deployment Manager',
                'external-doc-url': 'https://cloud.google.com/deployment-manager/',
                'logo': '/integration-logos/gcp/Google-Deployment-Manager.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google API Python Client',
                'external-doc-url': 'https://github.com/googleapis/google-api-python-client/',
                'logo': '/integration-logos/gcp/Google-API-Python-Client.png',
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Calendar',
                'external-doc-url': 'https://calendar.google.com/',
                'logo': '/integration-logos/gcp/Google-Calendar.png',
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Campaign Manager',
                'external-doc-url': 'https://developers.google.com/doubleclick-advertisers',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/marketing_platform/campaign_manager.rst'
                ],
                'logo': '/integration-logos/gcp/Google-Campaign-Manager.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud',
                'external-doc-url': 'https://cloud.google.com/',
                'logo': '/integration-logos/gcp/Google-Cloud.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Discovery API',
                'external-doc-url': 'https://developers.google.com/discovery/',
                'logo': '/integration-logos/gcp/Google-Cloud-Generic.png',
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Display&Video 360',
                'external-doc-url': 'https://marketingplatform.google.com/about/display-video-360/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/marketing_platform/display_video.rst'
                ],
                'logo': '/integration-logos/gcp/Google-Display-And-Video-360.png',
                'tags': ['gmp'],
            },
            {
                'integration-name': 'Google Drive',
                'external-doc-url': 'https://www.google.com/drive/',
                'logo': '/integration-logos/gcp/Google-Drive.png',
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Search Ads 360',
                'external-doc-url': 'https://marketingplatform.google.com/about/search-ads-360/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/marketing_platform/search_ads.rst'
                ],
                'logo': '/integration-logos/gcp/Google-Search-Ads360.png',
                'tags': ['gmp'],
            },
            {
                'integration-name': 'Google',
                'external-doc-url': 'https://developer.google.com/',
                'logo': '/integration-logos/gcp/Google.png',
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Spreadsheet',
                'external-doc-url': 'https://www.google.com/intl/en/sheets/about/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/suite/sheets.rst'],
                'logo': '/integration-logos/gcp/Google-Spreadsheet.png',
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Cloud Storage Transfer Service',
                'external-doc-url': 'https://cloud.google.com/storage-transfer-service/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/cloud_storage_transfer_service.rst'
                ],
                'logo': '/integration-logos/gcp/Cloud-Storage.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Kubernetes Engine',
                'external-doc-url': 'https://cloud.google.com/kubernetes-engine/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/kubernetes_engine.rst'
                ],
                'logo': '/integration-logos/gcp/Kubernetes-Engine.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Machine Learning Engine',
                'external-doc-url': 'https://cloud.google.com/vertex-ai/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/mlengine.rst'],
                'logo': '/integration-logos/gcp/AI-Platform.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Natural Language',
                'external-doc-url': 'https://cloud.google.com/natural-language/',
                'how-to-guide': [
                    '/docs/apache-airflow-providers-google/operators/cloud/natural_language.rst'
                ],
                'logo': '/integration-logos/gcp/Cloud-NLP.png',
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Cloud Workflows',
                'external-doc-url': 'https://cloud.google.com/workflows/',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/workflows.rst'],
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google LevelDB',
                'external-doc-url': 'https://github.com/google/leveldb/blob/master/doc/index.md',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/leveldb/leveldb.rst'],
                'tags': ['google'],
            },
            {
                'integration-name': 'Google Vertex AI',
                'external-doc-url': 'https://cloud.google.com/vertex-ai',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/vertex_ai.rst'],
                'tags': ['gcp'],
            },
            {
                'integration-name': 'Google Looker',
                'external-doc-url': 'https://cloud.google.com/looker',
                'logo': '/integration-logos/gcp/Cloud-Looker.png',
                'how-to-guide': ['/docs/apache-airflow-providers-google/operators/cloud/looker.rst'],
                'tags': ['gcp'],
            },
        ],
        'operators': [
            {
                'integration-name': 'Google Ads',
                'python-modules': ['airflow.providers.google.ads.operators.ads'],
            },
            {
                'integration-name': 'Google AutoML',
                'python-modules': ['airflow.providers.google.cloud.operators.automl'],
            },
            {
                'integration-name': 'Google BigQuery',
                'python-modules': ['airflow.providers.google.cloud.operators.bigquery'],
            },
            {
                'integration-name': 'Google BigQuery Data Transfer Service',
                'python-modules': ['airflow.providers.google.cloud.operators.bigquery_dts'],
            },
            {
                'integration-name': 'Google Bigtable',
                'python-modules': ['airflow.providers.google.cloud.operators.bigtable'],
            },
            {
                'integration-name': 'Google Cloud Build',
                'python-modules': ['airflow.providers.google.cloud.operators.cloud_build'],
            },
            {
                'integration-name': 'Google Cloud Composer',
                'python-modules': ['airflow.providers.google.cloud.operators.cloud_composer'],
            },
            {
                'integration-name': 'Google Cloud Memorystore',
                'python-modules': ['airflow.providers.google.cloud.operators.cloud_memorystore'],
            },
            {
                'integration-name': 'Google Cloud SQL',
                'python-modules': ['airflow.providers.google.cloud.operators.cloud_sql'],
            },
            {
                'integration-name': 'Google Cloud Storage Transfer Service',
                'python-modules': ['airflow.providers.google.cloud.operators.cloud_storage_transfer_service'],
            },
            {
                'integration-name': 'Google Compute Engine',
                'python-modules': ['airflow.providers.google.cloud.operators.compute'],
            },
            {
                'integration-name': 'Google Data Catalog',
                'python-modules': ['airflow.providers.google.cloud.operators.datacatalog'],
            },
            {
                'integration-name': 'Google Dataflow',
                'python-modules': ['airflow.providers.google.cloud.operators.dataflow'],
            },
            {
                'integration-name': 'Google Data Fusion',
                'python-modules': ['airflow.providers.google.cloud.operators.datafusion'],
            },
            {
                'integration-name': 'Google Dataplex',
                'python-modules': ['airflow.providers.google.cloud.operators.dataplex'],
            },
            {
                'integration-name': 'Google Dataprep',
                'python-modules': ['airflow.providers.google.cloud.operators.dataprep'],
            },
            {
                'integration-name': 'Google Dataproc Metastore',
                'python-modules': ['airflow.providers.google.cloud.operators.dataproc_metastore'],
            },
            {
                'integration-name': 'Google Dataproc',
                'python-modules': ['airflow.providers.google.cloud.operators.dataproc'],
            },
            {
                'integration-name': 'Google Datastore',
                'python-modules': ['airflow.providers.google.cloud.operators.datastore'],
            },
            {
                'integration-name': 'Google Cloud Data Loss Prevention (DLP)',
                'python-modules': ['airflow.providers.google.cloud.operators.dlp'],
            },
            {
                'integration-name': 'Google Cloud Functions',
                'python-modules': ['airflow.providers.google.cloud.operators.functions'],
            },
            {
                'integration-name': 'Google Cloud Storage (GCS)',
                'python-modules': ['airflow.providers.google.cloud.operators.gcs'],
            },
            {
                'integration-name': 'Google Kubernetes Engine',
                'python-modules': ['airflow.providers.google.cloud.operators.kubernetes_engine'],
            },
            {
                'integration-name': 'Google Cloud Life Sciences',
                'python-modules': ['airflow.providers.google.cloud.operators.life_sciences'],
            },
            {
                'integration-name': 'Google Machine Learning Engine',
                'python-modules': ['airflow.providers.google.cloud.operators.mlengine'],
            },
            {
                'integration-name': 'Google Cloud Natural Language',
                'python-modules': ['airflow.providers.google.cloud.operators.natural_language'],
            },
            {
                'integration-name': 'Google Cloud Pub/Sub',
                'python-modules': ['airflow.providers.google.cloud.operators.pubsub'],
            },
            {
                'integration-name': 'Google Cloud Spanner',
                'python-modules': ['airflow.providers.google.cloud.operators.spanner'],
            },
            {
                'integration-name': 'Google Cloud Speech-to-Text',
                'python-modules': ['airflow.providers.google.cloud.operators.speech_to_text'],
            },
            {
                'integration-name': 'Google Cloud Stackdriver',
                'python-modules': ['airflow.providers.google.cloud.operators.stackdriver'],
            },
            {
                'integration-name': 'Google Cloud Tasks',
                'python-modules': ['airflow.providers.google.cloud.operators.tasks'],
            },
            {
                'integration-name': 'Google Cloud Text-to-Speech',
                'python-modules': [
                    'airflow.providers.google.cloud.operators.text_to_speech',
                    'airflow.providers.google.cloud.operators.translate_speech',
                ],
            },
            {
                'integration-name': 'Google Cloud Translation',
                'python-modules': [
                    'airflow.providers.google.cloud.operators.translate',
                    'airflow.providers.google.cloud.operators.translate_speech',
                ],
            },
            {
                'integration-name': 'Google Cloud Video Intelligence',
                'python-modules': ['airflow.providers.google.cloud.operators.video_intelligence'],
            },
            {
                'integration-name': 'Google Cloud Vision',
                'python-modules': ['airflow.providers.google.cloud.operators.vision'],
            },
            {
                'integration-name': 'Google Cloud Workflows',
                'python-modules': ['airflow.providers.google.cloud.operators.workflows'],
            },
            {
                'integration-name': 'Google Cloud Firestore',
                'python-modules': ['airflow.providers.google.firebase.operators.firestore'],
            },
            {
                'integration-name': 'Google Analytics360',
                'python-modules': ['airflow.providers.google.marketing_platform.operators.analytics'],
            },
            {
                'integration-name': 'Google Campaign Manager',
                'python-modules': ['airflow.providers.google.marketing_platform.operators.campaign_manager'],
            },
            {
                'integration-name': 'Google Display&Video 360',
                'python-modules': ['airflow.providers.google.marketing_platform.operators.display_video'],
            },
            {
                'integration-name': 'Google Search Ads 360',
                'python-modules': ['airflow.providers.google.marketing_platform.operators.search_ads'],
            },
            {
                'integration-name': 'Google Spreadsheet',
                'python-modules': ['airflow.providers.google.suite.operators.sheets'],
            },
            {
                'integration-name': 'Google LevelDB',
                'python-modules': ['airflow.providers.google.leveldb.operators.leveldb'],
            },
            {
                'integration-name': 'Google Vertex AI',
                'python-modules': [
                    'airflow.providers.google.cloud.operators.vertex_ai.dataset',
                    'airflow.providers.google.cloud.operators.vertex_ai.custom_job',
                    'airflow.providers.google.cloud.operators.vertex_ai.auto_ml',
                    'airflow.providers.google.cloud.operators.vertex_ai.batch_prediction_job',
                    'airflow.providers.google.cloud.operators.vertex_ai.endpoint_service',
                    'airflow.providers.google.cloud.operators.vertex_ai.hyperparameter_tuning_job',
                    'airflow.providers.google.cloud.operators.vertex_ai.model_service',
                ],
            },
            {
                'integration-name': 'Google Looker',
                'python-modules': ['airflow.providers.google.cloud.operators.looker'],
            },
            {
                'integration-name': 'Google Cloud Dataform',
                'python-modules': ['airflow.providers.google.cloud.operators.dataform'],
            },
        ],
        'sensors': [
            {
                'integration-name': 'Google BigQuery',
                'python-modules': ['airflow.providers.google.cloud.sensors.bigquery'],
            },
            {
                'integration-name': 'Google BigQuery Data Transfer Service',
                'python-modules': ['airflow.providers.google.cloud.sensors.bigquery_dts'],
            },
            {
                'integration-name': 'Google Bigtable',
                'python-modules': ['airflow.providers.google.cloud.sensors.bigtable'],
            },
            {
                'integration-name': 'Google Cloud Storage Transfer Service',
                'python-modules': ['airflow.providers.google.cloud.sensors.cloud_storage_transfer_service'],
            },
            {
                'integration-name': 'Google Dataflow',
                'python-modules': ['airflow.providers.google.cloud.sensors.dataflow'],
            },
            {
                'integration-name': 'Google Data Fusion',
                'python-modules': ['airflow.providers.google.cloud.sensors.datafusion'],
            },
            {
                'integration-name': 'Google Dataplex',
                'python-modules': ['airflow.providers.google.cloud.sensors.dataplex'],
            },
            {
                'integration-name': 'Google Dataproc',
                'python-modules': ['airflow.providers.google.cloud.sensors.dataproc'],
            },
            {
                'integration-name': 'Google Cloud Storage (GCS)',
                'python-modules': ['airflow.providers.google.cloud.sensors.gcs'],
            },
            {
                'integration-name': 'Google Cloud Pub/Sub',
                'python-modules': ['airflow.providers.google.cloud.sensors.pubsub'],
            },
            {
                'integration-name': 'Google Cloud Workflows',
                'python-modules': ['airflow.providers.google.cloud.sensors.workflows'],
            },
            {
                'integration-name': 'Google Drive',
                'python-modules': ['airflow.providers.google.suite.sensors.drive'],
            },
            {
                'integration-name': 'Google Campaign Manager',
                'python-modules': ['airflow.providers.google.marketing_platform.sensors.campaign_manager'],
            },
            {
                'integration-name': 'Google Display&Video 360',
                'python-modules': ['airflow.providers.google.marketing_platform.sensors.display_video'],
            },
            {
                'integration-name': 'Google Search Ads 360',
                'python-modules': ['airflow.providers.google.marketing_platform.sensors.search_ads'],
            },
            {
                'integration-name': 'Google Looker',
                'python-modules': ['airflow.providers.google.cloud.sensors.looker'],
            },
            {
                'integration-name': 'Google Cloud Dataform',
                'python-modules': ['airflow.providers.google.cloud.sensors.dataform'],
            },
        ],
        'hooks': [
            {'integration-name': 'Google Ads', 'python-modules': ['airflow.providers.google.ads.hooks.ads']},
            {
                'integration-name': 'Google AutoML',
                'python-modules': ['airflow.providers.google.cloud.hooks.automl'],
            },
            {
                'integration-name': 'Google BigQuery',
                'python-modules': ['airflow.providers.google.cloud.hooks.bigquery'],
            },
            {
                'integration-name': 'Google BigQuery Data Transfer Service',
                'python-modules': ['airflow.providers.google.cloud.hooks.bigquery_dts'],
            },
            {
                'integration-name': 'Google Bigtable',
                'python-modules': ['airflow.providers.google.cloud.hooks.bigtable'],
            },
            {
                'integration-name': 'Google Cloud Build',
                'python-modules': ['airflow.providers.google.cloud.hooks.cloud_build'],
            },
            {
                'integration-name': 'Google Cloud Composer',
                'python-modules': ['airflow.providers.google.cloud.hooks.cloud_composer'],
            },
            {
                'integration-name': 'Google Cloud Memorystore',
                'python-modules': ['airflow.providers.google.cloud.hooks.cloud_memorystore'],
            },
            {
                'integration-name': 'Google Cloud SQL',
                'python-modules': ['airflow.providers.google.cloud.hooks.cloud_sql'],
            },
            {
                'integration-name': 'Google Cloud Storage Transfer Service',
                'python-modules': ['airflow.providers.google.cloud.hooks.cloud_storage_transfer_service'],
            },
            {
                'integration-name': 'Google Compute Engine',
                'python-modules': [
                    'airflow.providers.google.cloud.hooks.compute',
                    'airflow.providers.google.cloud.hooks.compute_ssh',
                ],
            },
            {
                'integration-name': 'Google Data Catalog',
                'python-modules': ['airflow.providers.google.cloud.hooks.datacatalog'],
            },
            {
                'integration-name': 'Google Dataflow',
                'python-modules': ['airflow.providers.google.cloud.hooks.dataflow'],
            },
            {
                'integration-name': 'Google Data Fusion',
                'python-modules': ['airflow.providers.google.cloud.hooks.datafusion'],
            },
            {
                'integration-name': 'Google Dataplex',
                'python-modules': ['airflow.providers.google.cloud.hooks.dataplex'],
            },
            {
                'integration-name': 'Google Dataprep',
                'python-modules': ['airflow.providers.google.cloud.hooks.dataprep'],
            },
            {
                'integration-name': 'Google Dataproc Metastore',
                'python-modules': ['airflow.providers.google.cloud.hooks.dataproc_metastore'],
            },
            {
                'integration-name': 'Google Dataproc',
                'python-modules': ['airflow.providers.google.cloud.hooks.dataproc'],
            },
            {
                'integration-name': 'Google Datastore',
                'python-modules': ['airflow.providers.google.cloud.hooks.datastore'],
            },
            {
                'integration-name': 'Google Cloud Data Loss Prevention (DLP)',
                'python-modules': ['airflow.providers.google.cloud.hooks.dlp'],
            },
            {
                'integration-name': 'Google Cloud Functions',
                'python-modules': ['airflow.providers.google.cloud.hooks.functions'],
            },
            {
                'integration-name': 'Google Cloud Storage (GCS)',
                'python-modules': ['airflow.providers.google.cloud.hooks.gcs'],
            },
            {
                'integration-name': 'Google Deployment Manager',
                'python-modules': ['airflow.providers.google.cloud.hooks.gdm'],
            },
            {
                'integration-name': 'Google Cloud Key Management Service (KMS)',
                'python-modules': ['airflow.providers.google.cloud.hooks.kms'],
            },
            {
                'integration-name': 'Google Kubernetes Engine',
                'python-modules': ['airflow.providers.google.cloud.hooks.kubernetes_engine'],
            },
            {
                'integration-name': 'Google Cloud Life Sciences',
                'python-modules': ['airflow.providers.google.cloud.hooks.life_sciences'],
            },
            {
                'integration-name': 'Google Machine Learning Engine',
                'python-modules': ['airflow.providers.google.cloud.hooks.mlengine'],
            },
            {
                'integration-name': 'Google Cloud Natural Language',
                'python-modules': ['airflow.providers.google.cloud.hooks.natural_language'],
            },
            {
                'integration-name': 'Google Cloud OS Login',
                'python-modules': ['airflow.providers.google.cloud.hooks.os_login'],
            },
            {
                'integration-name': 'Google Cloud Pub/Sub',
                'python-modules': ['airflow.providers.google.cloud.hooks.pubsub'],
            },
            {
                'integration-name': 'Google Cloud Secret Manager',
                'python-modules': ['airflow.providers.google.cloud.hooks.secret_manager'],
            },
            {
                'integration-name': 'Google Cloud Spanner',
                'python-modules': ['airflow.providers.google.cloud.hooks.spanner'],
            },
            {
                'integration-name': 'Google Cloud Speech-to-Text',
                'python-modules': ['airflow.providers.google.cloud.hooks.speech_to_text'],
            },
            {
                'integration-name': 'Google Cloud Stackdriver',
                'python-modules': ['airflow.providers.google.cloud.hooks.stackdriver'],
            },
            {
                'integration-name': 'Google Cloud Tasks',
                'python-modules': ['airflow.providers.google.cloud.hooks.tasks'],
            },
            {
                'integration-name': 'Google Cloud Text-to-Speech',
                'python-modules': ['airflow.providers.google.cloud.hooks.text_to_speech'],
            },
            {
                'integration-name': 'Google Cloud Translation',
                'python-modules': ['airflow.providers.google.cloud.hooks.translate'],
            },
            {
                'integration-name': 'Google Cloud Video Intelligence',
                'python-modules': ['airflow.providers.google.cloud.hooks.video_intelligence'],
            },
            {
                'integration-name': 'Google Cloud Vision',
                'python-modules': ['airflow.providers.google.cloud.hooks.vision'],
            },
            {
                'integration-name': 'Google Cloud Workflows',
                'python-modules': ['airflow.providers.google.cloud.hooks.workflows'],
            },
            {
                'integration-name': 'Google',
                'python-modules': ['airflow.providers.google.common.hooks.base_google'],
            },
            {
                'integration-name': 'Google Discovery API',
                'python-modules': ['airflow.providers.google.common.hooks.discovery_api'],
            },
            {
                'integration-name': 'Google Cloud Firestore',
                'python-modules': ['airflow.providers.google.firebase.hooks.firestore'],
            },
            {
                'integration-name': 'Google Analytics360',
                'python-modules': ['airflow.providers.google.marketing_platform.hooks.analytics'],
            },
            {
                'integration-name': 'Google Campaign Manager',
                'python-modules': ['airflow.providers.google.marketing_platform.hooks.campaign_manager'],
            },
            {
                'integration-name': 'Google Display&Video 360',
                'python-modules': ['airflow.providers.google.marketing_platform.hooks.display_video'],
            },
            {
                'integration-name': 'Google Search Ads 360',
                'python-modules': ['airflow.providers.google.marketing_platform.hooks.search_ads'],
            },
            {
                'integration-name': 'Google Calendar',
                'python-modules': ['airflow.providers.google.suite.hooks.calendar'],
            },
            {
                'integration-name': 'Google Drive',
                'python-modules': ['airflow.providers.google.suite.hooks.drive'],
            },
            {
                'integration-name': 'Google Spreadsheet',
                'python-modules': ['airflow.providers.google.suite.hooks.sheets'],
            },
            {
                'integration-name': 'Google LevelDB',
                'python-modules': ['airflow.providers.google.leveldb.hooks.leveldb'],
            },
            {
                'integration-name': 'Google Vertex AI',
                'python-modules': [
                    'airflow.providers.google.cloud.hooks.vertex_ai.dataset',
                    'airflow.providers.google.cloud.hooks.vertex_ai.custom_job',
                    'airflow.providers.google.cloud.hooks.vertex_ai.auto_ml',
                    'airflow.providers.google.cloud.hooks.vertex_ai.batch_prediction_job',
                    'airflow.providers.google.cloud.hooks.vertex_ai.endpoint_service',
                    'airflow.providers.google.cloud.hooks.vertex_ai.hyperparameter_tuning_job',
                    'airflow.providers.google.cloud.hooks.vertex_ai.model_service',
                ],
            },
            {
                'integration-name': 'Google Looker',
                'python-modules': ['airflow.providers.google.cloud.hooks.looker'],
            },
            {
                'integration-name': 'Google Cloud Dataform',
                'python-modules': ['airflow.providers.google.cloud.hooks.dataform'],
            },
        ],
        'transfers': [
            {
                'source-integration-name': 'Presto',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/presto_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.presto_to_gcs',
            },
            {
                'source-integration-name': 'Trino',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/trino_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.trino_to_gcs',
            },
            {
                'source-integration-name': 'SQL',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.sql_to_gcs',
            },
            {
                'source-integration-name': 'Google Cloud Storage (GCS)',
                'target-integration-name': 'Google Drive',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/gcs_to_gdrive.rst',
                'python-module': 'airflow.providers.google.suite.transfers.gcs_to_gdrive',
            },
            {
                'source-integration-name': 'Google Drive',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/gdrive_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.gdrive_to_gcs',
            },
            {
                'source-integration-name': 'Microsoft SQL Server (MSSQL)',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.mssql_to_gcs',
            },
            {
                'source-integration-name': 'Microsoft Azure FileShare',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': (
                    '/docs/apache-airflow-providers-google/operators/transfer/azure_fileshare_to_gcs.rst'
                ),
                'python-module': 'airflow.providers.google.cloud.transfers.azure_fileshare_to_gcs',
            },
            {
                'source-integration-name': 'Apache Cassandra',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.cassandra_to_gcs',
            },
            {
                'source-integration-name': 'Google Calendar',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': (
                    '/docs/apache-airflow-providers-google/operators/transfer/calendar_to_gcs.rst'
                ),
                'python-module': 'airflow.providers.google.cloud.transfers.calendar_to_gcs',
            },
            {
                'source-integration-name': 'Google Spreadsheet',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/sheets_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.sheets_to_gcs',
            },
            {
                'source-integration-name': 'Amazon Simple Storage Service (S3)',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/s3_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.s3_to_gcs',
            },
            {
                'source-integration-name': 'Google Cloud Storage (GCS)',
                'target-integration-name': 'SSH File Transfer Protocol (SFTP)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/gcs_to_sftp.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.gcs_to_sftp',
            },
            {
                'source-integration-name': 'PostgreSQL',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.postgres_to_gcs',
            },
            {
                'source-integration-name': 'Google BigQuery',
                'target-integration-name': 'MySQL',
                'python-module': 'airflow.providers.google.cloud.transfers.bigquery_to_mysql',
            },
            {
                'source-integration-name': 'Google BigQuery',
                'target-integration-name': 'Microsoft SQL Server (MSSQL)',
                'python-module': 'airflow.providers.google.cloud.transfers.bigquery_to_mssql',
            },
            {
                'source-integration-name': 'Google Cloud Storage (GCS)',
                'target-integration-name': 'Google BigQuery',
                'python-module': 'airflow.providers.google.cloud.transfers.gcs_to_bigquery',
            },
            {
                'source-integration-name': 'Google Cloud Storage (GCS)',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/gcs_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.gcs_to_gcs',
            },
            {
                'source-integration-name': 'Facebook Ads',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': (
                    '/docs/apache-airflow-providers-google/operators/transfer/facebook_ads_to_gcs.rst'
                ),
                'python-module': 'airflow.providers.google.cloud.transfers.facebook_ads_to_gcs',
            },
            {
                'source-integration-name': 'SSH File Transfer Protocol (SFTP)',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/sftp_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.sftp_to_gcs',
            },
            {
                'source-integration-name': 'Microsoft Azure Data Lake Storage',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.adls_to_gcs',
            },
            {
                'source-integration-name': 'Google BigQuery',
                'target-integration-name': 'Google BigQuery',
                'python-module': 'airflow.providers.google.cloud.transfers.bigquery_to_bigquery',
            },
            {
                'source-integration-name': 'MySQL',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.mysql_to_gcs',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/mysql_to_gcs.rst',
            },
            {
                'source-integration-name': 'Oracle',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.oracle_to_gcs',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/oracle_to_gcs.rst',
            },
            {
                'source-integration-name': 'Google Cloud Storage (GCS)',
                'target-integration-name': 'Google Spreadsheet',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/gcs_to_sheets.rst',
                'python-module': 'airflow.providers.google.suite.transfers.gcs_to_sheets',
            },
            {
                'source-integration-name': 'SQL',
                'target-integration-name': 'Google Spreadsheet',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/sql_to_sheets.rst',
                'python-module': 'airflow.providers.google.suite.transfers.sql_to_sheets',
            },
            {
                'source-integration-name': 'Local',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/local_to_gcs.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.local_to_gcs',
            },
            {
                'source-integration-name': 'Google BigQuery',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.bigquery_to_gcs',
            },
            {
                'source-integration-name': 'Google Cloud Storage (GCS)',
                'target-integration-name': 'Local',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/gcs_to_local.rst',
                'python-module': 'airflow.providers.google.cloud.transfers.gcs_to_local',
            },
            {
                'source-integration-name': 'Google Drive',
                'target-integration-name': 'Local',
                'how-to-guide': (
                    '/docs/apache-airflow-providers-google/operators/transfer/gdrive_to_local.rst'
                ),
                'python-module': 'airflow.providers.google.cloud.transfers.gdrive_to_local',
            },
            {
                'source-integration-name': 'Local',
                'target-integration-name': 'Google Drive',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/local_to_drive.rst',
                'python-module': 'airflow.providers.google.suite.transfers.local_to_drive',
            },
            {
                'source-integration-name': 'Salesforce',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'how-to-guide': (
                    '/docs/apache-airflow-providers-google/operators/transfer/salesforce_to_gcs.rst'
                ),
                'python-module': 'airflow.providers.google.cloud.transfers.salesforce_to_gcs',
            },
            {
                'source-integration-name': 'Google Ads',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.ads.transfers.ads_to_gcs',
            },
            {
                'source-integration-name': 'Microsoft SQL Server (MSSQL)',
                'target-integration-name': 'Google Cloud Storage (GCS)',
                'python-module': 'airflow.providers.google.cloud.transfers.mssql_to_gcs',
                'how-to-guide': '/docs/apache-airflow-providers-google/operators/transfer/mssql_to_gcs.rst',
            },
        ],
        'connection-types': [
            {
                'hook-class-name': 'airflow.providers.google.common.hooks.base_google.GoogleBaseHook',
                'connection-type': 'google_cloud_platform',
            },
            {
                'hook-class-name': 'airflow.providers.google.cloud.hooks.dataprep.GoogleDataprepHook',
                'connection-type': 'dataprep',
            },
            {
                'hook-class-name': 'airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook',
                'connection-type': 'gcpcloudsql',
            },
            {
                'hook-class-name': 'airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook',
                'connection-type': 'gcpcloudsqldb',
            },
            {
                'hook-class-name': 'airflow.providers.google.cloud.hooks.bigquery.BigQueryHook',
                'connection-type': 'gcpbigquery',
            },
            {
                'hook-class-name': 'airflow.providers.google.cloud.hooks.compute_ssh.ComputeEngineSSHHook',
                'connection-type': 'gcpssh',
            },
            {
                'hook-class-name': 'airflow.providers.google.leveldb.hooks.leveldb.LevelDBHook',
                'connection-type': 'leveldb',
            },
        ],
        'extra-links': [
            'airflow.providers.google.cloud.operators.bigquery.BigQueryConsoleLink',
            'airflow.providers.google.cloud.operators.bigquery.BigQueryConsoleIndexableLink',
            'airflow.providers.google.cloud.operators.mlengine.AIPlatformConsoleLink',
            'airflow.providers.google.cloud.links.dataform.DataformWorkflowInvocationLink',
            'airflow.providers.google.cloud.operators.datafusion.DataFusionInstanceLink',
            'airflow.providers.google.cloud.operators.datafusion.DataFusionPipelineLink',
            'airflow.providers.google.cloud.operators.datafusion.DataFusionPipelinesLink',
            'airflow.providers.google.cloud.links.cloud_sql.CloudSQLInstanceLink',
            'airflow.providers.google.cloud.links.cloud_sql.CloudSQLInstanceDatabaseLink',
            'airflow.providers.google.cloud.links.dataplex.DataplexTaskLink',
            'airflow.providers.google.cloud.links.dataplex.DataplexTasksLink',
            'airflow.providers.google.cloud.links.bigquery.BigQueryDatasetLink',
            'airflow.providers.google.cloud.links.bigquery.BigQueryTableLink',
            'airflow.providers.google.cloud.links.bigquery_dts.BigQueryDataTransferConfigLink',
            'airflow.providers.google.cloud.links.cloud_tasks.CloudTasksQueueLink',
            'airflow.providers.google.cloud.links.cloud_tasks.CloudTasksLink',
            'airflow.providers.google.cloud.links.datacatalog.DataCatalogEntryGroupLink',
            'airflow.providers.google.cloud.links.datacatalog.DataCatalogEntryLink',
            'airflow.providers.google.cloud.links.datacatalog.DataCatalogTagTemplateLink',
            'airflow.providers.google.cloud.links.dataproc.DataprocLink',
            'airflow.providers.google.cloud.links.dataproc.DataprocListLink',
            'airflow.providers.google.cloud.operators.dataproc_metastore.DataprocMetastoreDetailedLink',
            'airflow.providers.google.cloud.operators.dataproc_metastore.DataprocMetastoreLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIModelLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIModelListLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIModelExportLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAITrainingLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAITrainingPipelinesLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIDatasetLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIDatasetListLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIHyperparameterTuningJobListLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIBatchPredictionJobLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIBatchPredictionJobListLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIEndpointLink',
            'airflow.providers.google.cloud.links.vertex_ai.VertexAIEndpointListLink',
            'airflow.providers.google.cloud.links.workflows.WorkflowsWorkflowDetailsLink',
            'airflow.providers.google.cloud.links.workflows.WorkflowsListOfWorkflowsLink',
            'airflow.providers.google.cloud.links.workflows.WorkflowsExecutionLink',
            'airflow.providers.google.cloud.operators.cloud_composer.CloudComposerEnvironmentLink',
            'airflow.providers.google.cloud.operators.cloud_composer.CloudComposerEnvironmentsLink',
            'airflow.providers.google.cloud.links.dataflow.DataflowJobLink',
            'airflow.providers.google.cloud.links.datastore.CloudDatastoreImportExportLink',
            'airflow.providers.google.cloud.links.datastore.CloudDatastoreEntitiesLink',
            'airflow.providers.google.cloud.links.bigtable.BigtableInstanceLink',
            'airflow.providers.google.cloud.links.bigtable.BigtableClusterLink',
            'airflow.providers.google.cloud.links.bigtable.BigtableTablesLink',
            'airflow.providers.google.cloud.links.spanner.SpannerDatabaseLink',
            'airflow.providers.google.cloud.links.spanner.SpannerInstanceLink',
            'airflow.providers.google.cloud.links.stackdriver.StackdriverNotificationsLink',
            'airflow.providers.google.cloud.links.stackdriver.StackdriverPoliciesLink',
            'airflow.providers.google.cloud.links.kubernetes_engine.KubernetesEngineClusterLink',
            'airflow.providers.google.cloud.links.kubernetes_engine.KubernetesEnginePodLink',
            'airflow.providers.google.cloud.links.pubsub.PubSubSubscriptionLink',
            'airflow.providers.google.cloud.links.pubsub.PubSubTopicLink',
            'airflow.providers.google.cloud.links.cloud_memorystore.MemcachedInstanceDetailsLink',
            'airflow.providers.google.cloud.links.cloud_memorystore.MemcachedInstanceListLink',
            'airflow.providers.google.cloud.links.cloud_memorystore.RedisInstanceDetailsLink',
            'airflow.providers.google.cloud.links.cloud_memorystore.RedisInstanceListLink',
            'airflow.providers.google.common.links.storage.StorageLink',
            'airflow.providers.google.common.links.storage.FileDetailsLink',
        ],
        'additional-extras': [
            {'name': 'apache.beam', 'dependencies': ['apache-beam[gcp]']},
            {'name': 'leveldb', 'dependencies': ['plyvel']},
            {'name': 'oracle', 'dependencies': ['apache-airflow-providers-oracle>=3.1.0']},
            {'name': 'facebook', 'dependencies': ['apache-airflow-providers-facebook>=2.2.0']},
            {'name': 'amazon', 'dependencies': ['apache-airflow-providers-amazon>=2.6.0']},
        ],
        'secrets-backends': [
            'airflow.providers.google.cloud.secrets.secret_manager.CloudSecretManagerBackend'
        ],
        'auth-backends': ['airflow.providers.google.common.auth_backend.google_openid'],
        'logging': [
            'airflow.providers.google.cloud.log.gcs_task_handler.GCSTaskHandler',
            'airflow.providers.google.cloud.log.stackdriver_task_handler.StackdriverTaskHandler',
        ],
    }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Compute Engine operators."""

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Union

from googleapiclient.errors import HttpError
from json_merge_patch import merge

from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.providers.google.cloud.hooks.compute import ComputeEngineHook
from airflow.providers.google.cloud.utils.field_sanitizer import GcpBodyFieldSanitizer
from airflow.providers.google.cloud.utils.field_validator import GcpBodyFieldValidator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class ComputeEngineBaseOperator(BaseOperator):
    """Abstract base operator for Google Compute Engine operators to inherit from."""

    def __init__(
        self,
        *,
        zone: str,
        resource_id: str,
        project_id: Optional[str] = None,
        gcp_conn_id: str = 'google_cloud_default',
        api_version: str = 'v1',
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:
        self.project_id = project_id
        self.zone = zone
        self.resource_id = resource_id
        self.gcp_conn_id = gcp_conn_id
        self.api_version = api_version
        self.impersonation_chain = impersonation_chain
        self._validate_inputs()
        super().__init__(**kwargs)

    def _validate_inputs(self) -> None:
        if self.project_id == '':
            raise AirflowException("The required parameter 'project_id' is missing")
        if not self.zone:
            raise AirflowException("The required parameter 'zone' is missing")
        if not self.resource_id:
            raise AirflowException("The required parameter 'resource_id' is missing")

    def execute(self, context: 'Context'):
        pass


class ComputeEngineStartInstanceOperator(ComputeEngineBaseOperator):
    """
    Starts an instance in Google Compute Engine.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:ComputeEngineStartInstanceOperator`

    :param zone: Google Cloud zone where the instance exists.
    :param resource_id: Name of the Compute Engine instance resource.
    :param project_id: Optional, Google Cloud Project ID where the Compute
        Engine Instance exists. If set to None or missing, the default project_id from the Google Cloud
        connection is used.
    :param gcp_conn_id: Optional, The connection ID used to connect to Google Cloud.
        Defaults to 'google_cloud_default'.
    :param api_version: Optional, API version used (for example v1 - or beta). Defaults
        to v1.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    # [START gce_instance_start_template_fields]
    template_fields: Sequence[str] = (
        'project_id',
        'zone',
        'resource_id',
        'gcp_conn_id',
        'api_version',
        'impersonation_chain',
    )
    # [END gce_instance_start_template_fields]

    def execute(self, context: 'Context') -> None:
        hook = ComputeEngineHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        return hook.start_instance(zone=self.zone, resource_id=self.resource_id, project_id=self.project_id)


class ComputeEngineStopInstanceOperator(ComputeEngineBaseOperator):
    """
    Stops an instance in Google Compute Engine.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:ComputeEngineStopInstanceOperator`

    :param zone: Google Cloud zone where the instance exists.
    :param resource_id: Name of the Compute Engine instance resource.
    :param project_id: Optional, Google Cloud Project ID where the Compute
        Engine Instance exists. If set to None or missing, the default project_id from the Google Cloud
        connection is used.
    :param gcp_conn_id: Optional, The connection ID used to connect to Google Cloud.
        Defaults to 'google_cloud_default'.
    :param api_version: Optional, API version used (for example v1 - or beta). Defaults
        to v1.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    # [START gce_instance_stop_template_fields]
    template_fields: Sequence[str] = (
        'project_id',
        'zone',
        'resource_id',
        'gcp_conn_id',
        'api_version',
        'impersonation_chain',
    )
    # [END gce_instance_stop_template_fields]

    def execute(self, context: 'Context') -> None:
        hook = ComputeEngineHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        hook.stop_instance(zone=self.zone, resource_id=self.resource_id, project_id=self.project_id)


SET_MACHINE_TYPE_VALIDATION_SPECIFICATION = [
    dict(name="machineType", regexp="^.+$"),
]


class ComputeEngineSetMachineTypeOperator(ComputeEngineBaseOperator):
    """
    Changes the machine type for a stopped instance to the machine type specified in
        the request.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:ComputeEngineSetMachineTypeOperator`

    :param zone: Google Cloud zone where the instance exists.
    :param resource_id: Name of the Compute Engine instance resource.
    :param body: Body required by the Compute Engine setMachineType API, as described in
        https://cloud.google.com/compute/docs/reference/rest/v1/instances/setMachineType#request-body
    :param project_id: Optional, Google Cloud Project ID where the Compute
        Engine Instance exists. If set to None or missing, the default project_id from the Google Cloud
        connection is used.
    :param gcp_conn_id: Optional, The connection ID used to connect to Google Cloud.
        Defaults to 'google_cloud_default'.
    :param api_version: Optional, API version used (for example v1 - or beta). Defaults
        to v1.
    :param validate_body: Optional, If set to False, body validation is not performed.
        Defaults to False.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    # [START gce_instance_set_machine_type_template_fields]
    template_fields: Sequence[str] = (
        'project_id',
        'zone',
        'resource_id',
        'body',
        'gcp_conn_id',
        'api_version',
        'impersonation_chain',
    )
    # [END gce_instance_set_machine_type_template_fields]

    def __init__(
        self,
        *,
        zone: str,
        resource_id: str,
        body: dict,
        project_id: Optional[str] = None,
        gcp_conn_id: str = 'google_cloud_default',
        api_version: str = 'v1',
        validate_body: bool = True,
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:
        self.body = body
        self._field_validator = None  # type: Optional[GcpBodyFieldValidator]
        if validate_body:
            self._field_validator = GcpBodyFieldValidator(
                SET_MACHINE_TYPE_VALIDATION_SPECIFICATION, api_version=api_version
            )
        super().__init__(
            project_id=project_id,
            zone=zone,
            resource_id=resource_id,
            gcp_conn_id=gcp_conn_id,
            api_version=api_version,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )

    def _validate_all_body_fields(self) -> None:
        if self._field_validator:
            self._field_validator.validate(self.body)

    def execute(self, context: 'Context') -> None:
        hook = ComputeEngineHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self._validate_all_body_fields()
        return hook.set_machine_type(
            zone=self.zone, resource_id=self.resource_id, body=self.body, project_id=self.project_id
        )


GCE_INSTANCE_TEMPLATE_VALIDATION_PATCH_SPECIFICATION = [
    dict(name="name", regexp="^.+$"),
    dict(name="description", optional=True),
    dict(
        name="properties",
        type='dict',
        optional=True,
        fields=[
            dict(name="description", optional=True),
            dict(name="tags", optional=True, fields=[dict(name="items", optional=True)]),
            dict(name="machineType", optional=True),
            dict(name="canIpForward", optional=True),
            dict(name="networkInterfaces", optional=True),  # not validating deeper
            dict(name="disks", optional=True),  # not validating the array deeper
            dict(
                name="metadata",
                optional=True,
                fields=[
                    dict(name="fingerprint", optional=True),
                    dict(name="items", optional=True),
                    dict(name="kind", optional=True),
                ],
            ),
            dict(name="serviceAccounts", optional=True),  # not validating deeper
            dict(
                name="scheduling",
                optional=True,
                fields=[
                    dict(name="onHostMaintenance", optional=True),
                    dict(name="automaticRestart", optional=True),
                    dict(name="preemptible", optional=True),
                    dict(name="nodeAffinities", optional=True),  # not validating deeper
                ],
            ),
            dict(name="labels", optional=True),
            dict(name="guestAccelerators", optional=True),  # not validating deeper
            dict(name="minCpuPlatform", optional=True),
        ],
    ),
]  # type: List[Dict[str, Any]]

GCE_INSTANCE_TEMPLATE_FIELDS_TO_SANITIZE = [
    "kind",
    "id",
    "name",
    "creationTimestamp",
    "properties.disks.sha256",
    "properties.disks.kind",
    "properties.disks.sourceImageEncryptionKey.sha256",
    "properties.disks.index",
    "properties.disks.licenses",
    "properties.networkInterfaces.kind",
    "properties.networkInterfaces.accessConfigs.kind",
    "properties.networkInterfaces.name",
    "properties.metadata.kind",
    "selfLink",
]


class ComputeEngineCopyInstanceTemplateOperator(ComputeEngineBaseOperator):
    """
    Copies the instance template, applying specified changes.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:ComputeEngineCopyInstanceTemplateOperator`

    :param resource_id: Name of the Instance Template
    :param body_patch: Patch to the body of instanceTemplates object following rfc7386
        PATCH semantics. The body_patch content follows
        https://cloud.google.com/compute/docs/reference/rest/v1/instanceTemplates
        Name field is required as we need to rename the template,
        all the other fields are optional. It is important to follow PATCH semantics
        - arrays are replaced fully, so if you need to update an array you should
        provide the whole target array as patch element.
    :param project_id: Optional, Google Cloud Project ID where the Compute
        Engine Instance exists. If set to None or missing, the default project_id from the Google Cloud
        connection is used.
    :param request_id: Optional, unique request_id that you might add to achieve
        full idempotence (for example when client call times out repeating the request
        with the same request id will not create a new instance template again).
        It should be in UUID format as defined in RFC 4122.
    :param gcp_conn_id: Optional, The connection ID used to connect to Google Cloud.
        Defaults to 'google_cloud_default'.
    :param api_version: Optional, API version used (for example v1 - or beta). Defaults
        to v1.
    :param validate_body: Optional, If set to False, body validation is not performed.
        Defaults to False.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    # [START gce_instance_template_copy_operator_template_fields]
    template_fields: Sequence[str] = (
        'project_id',
        'resource_id',
        'request_id',
        'gcp_conn_id',
        'api_version',
        'impersonation_chain',
    )
    # [END gce_instance_template_copy_operator_template_fields]

    def __init__(
        self,
        *,
        resource_id: str,
        body_patch: dict,
        project_id: Optional[str] = None,
        request_id: Optional[str] = None,
        gcp_conn_id: str = 'google_cloud_default',
        api_version: str = 'v1',
        validate_body: bool = True,
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:
        self.body_patch = body_patch
        self.request_id = request_id
        self._field_validator = None  # Optional[GcpBodyFieldValidator]
        if 'name' not in self.body_patch:
            raise AirflowException(
                f"The body '{body_patch}' should contain at least name for the new operator "
                f"in the 'name' field"
            )
        if validate_body:
            self._field_validator = GcpBodyFieldValidator(
                GCE_INSTANCE_TEMPLATE_VALIDATION_PATCH_SPECIFICATION, api_version=api_version
            )
        self._field_sanitizer = GcpBodyFieldSanitizer(GCE_INSTANCE_TEMPLATE_FIELDS_TO_SANITIZE)
        super().__init__(
            project_id=project_id,
            zone='global',
            resource_id=resource_id,
            gcp_conn_id=gcp_conn_id,
            api_version=api_version,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )

    def _validate_all_body_fields(self) -> None:
        if self._field_validator:
            self._field_validator.validate(self.body_patch)

    def execute(self, context: 'Context') -> dict:
        hook = ComputeEngineHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self._validate_all_body_fields()
        try:
            # Idempotence check (sort of) - we want to check if the new template
            # is already created and if is, then we assume it was created by previous run
            # of CopyTemplate operator - we do not check if content of the template
            # is as expected. Templates are immutable so we cannot update it anyway
            # and deleting/recreating is not worth the hassle especially
            # that we cannot delete template if it is already used in some Instance
            # Group Manager. We assume success if the template is simply present
            existing_template = hook.get_instance_template(
                resource_id=self.body_patch['name'], project_id=self.project_id
            )
            self.log.info(
                "The %s template already existed. It was likely created by previous run of the operator. "
                "Assuming success.",
                existing_template,
            )
            return existing_template
        except HttpError as e:
            # We actually expect to get 404 / Not Found here as the template should
            # not yet exist
            if not e.resp.status == 404:
                raise e
        old_body = hook.get_instance_template(resource_id=self.resource_id, project_id=self.project_id)
        new_body = deepcopy(old_body)
        self._field_sanitizer.sanitize(new_body)
        new_body = merge(new_body, self.body_patch)
        self.log.info("Calling insert instance template with updated body: %s", new_body)
        hook.insert_instance_template(body=new_body, request_id=self.request_id, project_id=self.project_id)
        return hook.get_instance_template(resource_id=self.body_patch['name'], project_id=self.project_id)


class ComputeEngineInstanceGroupUpdateManagerTemplateOperator(ComputeEngineBaseOperator):
    """
    Patches the Instance Group Manager, replacing source template URL with the
    destination one. API V1 does not have update/patch operations for Instance
    Group Manager, so you must use beta or newer API version. Beta is the default.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:ComputeEngineInstanceGroupUpdateManagerTemplateOperator`

    :param resource_id: Name of the Instance Group Manager
    :param zone: Google Cloud zone where the Instance Group Manager exists.
    :param source_template: URL of the template to replace.
    :param destination_template: URL of the target template.
    :param project_id: Optional, Google Cloud Project ID where the Compute
        Engine Instance exists. If set to None or missing, the default project_id from the Google Cloud
        connection is used.
    :param request_id: Optional, unique request_id that you might add to achieve
        full idempotence (for example when client call times out repeating the request
        with the same request id will not create a new instance template again).
        It should be in UUID format as defined in RFC 4122.
    :param gcp_conn_id: Optional, The connection ID used to connect to Google Cloud.
        Defaults to 'google_cloud_default'.
    :param api_version: Optional, API version used (for example v1 - or beta). Defaults
        to v1.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    # [START gce_igm_update_template_operator_template_fields]
    template_fields: Sequence[str] = (
        'project_id',
        'resource_id',
        'zone',
        'request_id',
        'source_template',
        'destination_template',
        'gcp_conn_id',
        'api_version',
        'impersonation_chain',
    )
    # [END gce_igm_update_template_operator_template_fields]

    def __init__(
        self,
        *,
        resource_id: str,
        zone: str,
        source_template: str,
        destination_template: str,
        project_id: Optional[str] = None,
        update_policy: Optional[Dict[str, Any]] = None,
        request_id: Optional[str] = None,
        gcp_conn_id: str = 'google_cloud_default',
        api_version='beta',
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:
        self.zone = zone
        self.source_template = source_template
        self.destination_template = destination_template
        self.request_id = request_id
        self.update_policy = update_policy
        self._change_performed = False
        if api_version == 'v1':
            raise AirflowException(
                "Api version v1 does not have update/patch "
                "operations for Instance Group Managers. Use beta"
                " api version or above"
            )
        super().__init__(
            project_id=project_id,
            zone=self.zone,
            resource_id=resource_id,
            gcp_conn_id=gcp_conn_id,
            api_version=api_version,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )

    def _possibly_replace_template(self, dictionary: dict) -> None:
        if dictionary.get('instanceTemplate') == self.source_template:
            dictionary['instanceTemplate'] = self.destination_template
            self._change_performed = True

    def execute(self, context: 'Context') -> Optional[bool]:
        hook = ComputeEngineHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        old_instance_group_manager = hook.get_instance_group_manager(
            zone=self.zone, resource_id=self.resource_id, project_id=self.project_id
        )
        patch_body = {}
        if 'versions' in old_instance_group_manager:
            patch_body['versions'] = old_instance_group_manager['versions']
        if 'instanceTemplate' in old_instance_group_manager:
            patch_body['instanceTemplate'] = old_instance_group_manager['instanceTemplate']
        if self.update_policy:
            patch_body['updatePolicy'] = self.update_policy
        self._possibly_replace_template(patch_body)
        if 'versions' in patch_body:
            for version in patch_body['versions']:
                self._possibly_replace_template(version)
        if self._change_performed or self.update_policy:
            self.log.info("Calling patch instance template with updated body: %s", patch_body)
            return hook.patch_instance_group_manager(
                zone=self.zone,
                resource_id=self.resource_id,
                body=patch_body,
                request_id=self.request_id,
                project_id=self.project_id,
            )
        else:
            # Idempotence achieved
            return True



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os
from typing import Collection, Optional

# not sure why but mypy complains on missing `storage` but it is clearly there and is importable
from google.cloud import storage  # type: ignore[attr-defined]

from airflow.compat.functools import cached_property
from airflow.providers.google.cloud.utils.credentials_provider import get_credentials_and_project_id
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.utils.log.file_task_handler import FileTaskHandler
from airflow.utils.log.logging_mixin import LoggingMixin

_DEFAULT_SCOPESS = frozenset(
    [
        "https://www.googleapis.com/auth/devstorage.read_write",
    ]
)


class GCSTaskHandler(FileTaskHandler, LoggingMixin):
    """
    GCSTaskHandler is a python log handler that handles and reads
    task instance logs. It extends airflow FileTaskHandler and
    uploads to and reads from GCS remote storage. Upon log reading
    failure, it reads from host machine's local disk.

    :param base_log_folder: Base log folder to place logs.
    :param gcs_log_folder: Path to a remote location where logs will be saved. It must have the prefix
        ``gs://``. For example: ``gs://bucket/remote/log/location``
    :param filename_template: template filename string
    :param gcp_key_path: Path to Google Cloud Service Account file (JSON). Mutually exclusive with
        gcp_keyfile_dict.
        If omitted, authorization based on `the Application Default Credentials
        <https://cloud.google.com/docs/authentication/production#finding_credentials_automatically>`__ will
        be used.
    :param gcp_keyfile_dict: Dictionary of keyfile parameters. Mutually exclusive with gcp_key_path.
    :param gcp_scopes: Comma-separated string containing OAuth2 scopes
    :param project_id: Project ID to read the secrets from. If not passed, the project ID from credentials
        will be used.
    """

    def __init__(
        self,
        *,
        base_log_folder: str,
        gcs_log_folder: str,
        filename_template: Optional[str] = None,
        gcp_key_path: Optional[str] = None,
        gcp_keyfile_dict: Optional[dict] = None,
        gcp_scopes: Optional[Collection[str]] = _DEFAULT_SCOPESS,
        project_id: Optional[str] = None,
    ):
        super().__init__(base_log_folder, filename_template)
        self.remote_base = gcs_log_folder
        self.log_relative_path = ''
        self._hook = None
        self.closed = False
        self.upload_on_close = True
        self.gcp_key_path = gcp_key_path
        self.gcp_keyfile_dict = gcp_keyfile_dict
        self.scopes = gcp_scopes
        self.project_id = project_id

    @cached_property
    def client(self) -> storage.Client:
        """Returns GCS Client."""
        credentials, project_id = get_credentials_and_project_id(
            key_path=self.gcp_key_path,
            keyfile_dict=self.gcp_keyfile_dict,
            scopes=self.scopes,
            disable_logging=True,
        )
        return storage.Client(
            credentials=credentials,
            client_info=CLIENT_INFO,
            project=self.project_id if self.project_id else project_id,
        )

    def set_context(self, ti):
        super().set_context(ti)
        # Log relative path is used to construct local and remote
        # log path to upload log files into GCS and read from the
        # remote location.
        self.log_relative_path = self._render_filename(ti, ti.try_number)
        self.upload_on_close = not ti.raw

    def close(self):
        """Close and upload local log file to remote storage GCS."""
        # When application exit, system shuts down all handlers by
        # calling close method. Here we check if logger is already
        # closed to prevent uploading the log to remote storage multiple
        # times when `logging.shutdown` is called.
        if self.closed:
            return

        super().close()

        if not self.upload_on_close:
            return

        local_loc = os.path.join(self.local_base, self.log_relative_path)
        remote_loc = os.path.join(self.remote_base, self.log_relative_path)
        if os.path.exists(local_loc):
            # read log and remove old logs to get just the latest additions
            with open(local_loc) as logfile:
                log = logfile.read()
            self.gcs_write(log, remote_loc)

        # Mark closed so we don't double write if close is called twice
        self.closed = True

    def _read(self, ti, try_number, metadata=None):
        """
        Read logs of given task instance and try_number from GCS.
        If failed, read the log from task instance host machine.

        :param ti: task instance object
        :param try_number: task instance try_number to read logs from
        :param metadata: log metadata,
                         can be used for steaming log reading and auto-tailing.
        """
        # Explicitly getting log relative path is necessary as the given
        # task instance might be different than task instance passed in
        # in set_context method.
        log_relative_path = self._render_filename(ti, try_number)
        remote_loc = os.path.join(self.remote_base, log_relative_path)

        try:
            blob = storage.Blob.from_string(remote_loc, self.client)
            remote_log = blob.download_as_bytes().decode()
            log = f'*** Reading remote log from {remote_loc}.\n{remote_log}\n'
            return log, {'end_of_log': True}
        except Exception as e:
            log = f'*** Unable to read remote log from {remote_loc}\n*** {str(e)}\n\n'
            self.log.error(log)
            local_log, metadata = super()._read(ti, try_number)
            log += local_log
            return log, metadata

    def gcs_write(self, log, remote_log_location):
        """
        Writes the log to the remote_log_location. Fails silently if no log
        was created.

        :param log: the log to write to the remote_log_location
        :param remote_log_location: the log's location in remote storage
        """
        try:
            blob = storage.Blob.from_string(remote_log_location, self.client)
            old_log = blob.download_as_bytes().decode()
            log = '\n'.join([old_log, log]) if old_log else log
        except Exception as e:
            if not hasattr(e, 'resp') or e.resp.get('status') != '404':
                log = f'*** Previous log discarded: {str(e)}\n\n' + log
                self.log.info("Previous log discarded: %s", e)

        try:
            blob = storage.Blob.from_string(remote_log_location, self.client)
            blob.upload_from_string(log, content_type="text/plain")
        except Exception as e:
            self.log.error('Could not write logs to %s: %s', remote_log_location, e)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import TYPE_CHECKING

from airflow.providers.google.cloud.links.base import BaseGoogleLink

if TYPE_CHECKING:
    from airflow.utils.context import Context

BASE_LINK = "https://console.cloud.google.com"
VERTEX_AI_BASE_LINK = BASE_LINK + "/vertex-ai"
VERTEX_AI_MODEL_LINK = (
    VERTEX_AI_BASE_LINK + "/locations/{region}/models/{model_id}/deploy?project={project_id}"
)
VERTEX_AI_MODEL_LIST_LINK = VERTEX_AI_BASE_LINK + "/models?project={project_id}"
VERTEX_AI_MODEL_EXPORT_LINK = (
    BASE_LINK + "/storage/browser/{bucket_name}/model-{model_id}?project={project_id}"
)
VERTEX_AI_TRAINING_LINK = (
    VERTEX_AI_BASE_LINK + "/locations/{region}/training/{training_id}/cpu?project={project_id}"
)
VERTEX_AI_TRAINING_PIPELINES_LINK = VERTEX_AI_BASE_LINK + "/training/training-pipelines?project={project_id}"
VERTEX_AI_DATASET_LINK = (
    VERTEX_AI_BASE_LINK + "/locations/{region}/datasets/{dataset_id}/analyze?project={project_id}"
)
VERTEX_AI_DATASET_LIST_LINK = VERTEX_AI_BASE_LINK + "/datasets?project={project_id}"
VERTEX_AI_HYPERPARAMETER_TUNING_JOB_LIST_LINK = (
    VERTEX_AI_BASE_LINK + "/training/hyperparameter-tuning-jobs?project={project_id}"
)
VERTEX_AI_BATCH_PREDICTION_JOB_LINK = (
    VERTEX_AI_BASE_LINK
    + "/locations/{region}/batch-predictions/{batch_prediction_job_id}?project={project_id}"
)
VERTEX_AI_BATCH_PREDICTION_JOB_LIST_LINK = VERTEX_AI_BASE_LINK + "/batch-predictions?project={project_id}"
VERTEX_AI_ENDPOINT_LINK = (
    VERTEX_AI_BASE_LINK + "/locations/{region}/endpoints/{endpoint_id}?project={project_id}"
)
VERTEX_AI_ENDPOINT_LIST_LINK = VERTEX_AI_BASE_LINK + "/endpoints?project={project_id}"


class VertexAIModelLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Model link"""

    name = "Vertex AI Model"
    key = "model_conf"
    format_str = VERTEX_AI_MODEL_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
        model_id: str,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIModelLink.key,
            value={
                "model_id": model_id,
                "region": task_instance.region,
                "project_id": task_instance.project_id,
            },
        )


class VertexAIModelListLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Models Link"""

    name = "Model List"
    key = "models_conf"
    format_str = VERTEX_AI_MODEL_LIST_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIModelListLink.key,
            value={
                "project_id": task_instance.project_id,
            },
        )


class VertexAIModelExportLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Model Export Link"""

    name = "Export Model"
    key = "export_conf"
    format_str = VERTEX_AI_MODEL_EXPORT_LINK

    @staticmethod
    def extract_bucket_name(config):
        """Returns bucket name from output configuration."""
        return config["artifact_destination"]["output_uri_prefix"].rpartition("gs://")[-1]

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIModelExportLink.key,
            value={
                "project_id": task_instance.project_id,
                "model_id": task_instance.model_id,
                "bucket_name": VertexAIModelExportLink.extract_bucket_name(task_instance.output_config),
            },
        )


class VertexAITrainingLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Training link"""

    name = "Vertex AI Training"
    key = "training_conf"
    format_str = VERTEX_AI_TRAINING_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
        training_id: str,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAITrainingLink.key,
            value={
                "training_id": training_id,
                "region": task_instance.region,
                "project_id": task_instance.project_id,
            },
        )


class VertexAITrainingPipelinesLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Training Pipelines link"""

    name = "Vertex AI Training Pipelines"
    key = "pipelines_conf"
    format_str = VERTEX_AI_TRAINING_PIPELINES_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAITrainingPipelinesLink.key,
            value={
                "project_id": task_instance.project_id,
            },
        )


class VertexAIDatasetLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Dataset link"""

    name = "Dataset"
    key = "dataset_conf"
    format_str = VERTEX_AI_DATASET_LINK

    @staticmethod
    def persist(context: "Context", task_instance, dataset_id: str):
        task_instance.xcom_push(
            context=context,
            key=VertexAIDatasetLink.key,
            value={
                "dataset_id": dataset_id,
                "region": task_instance.region,
                "project_id": task_instance.project_id,
            },
        )


class VertexAIDatasetListLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Datasets Link"""

    name = "Dataset List"
    key = "datasets_conf"
    format_str = VERTEX_AI_DATASET_LIST_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIDatasetListLink.key,
            value={
                "project_id": task_instance.project_id,
            },
        )


class VertexAIHyperparameterTuningJobListLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI HyperparameterTuningJobs Link"""

    name = "Hyperparameter Tuning Job List"
    key = "hyperparameter_tuning_jobs_conf"
    format_str = VERTEX_AI_HYPERPARAMETER_TUNING_JOB_LIST_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIHyperparameterTuningJobListLink.key,
            value={
                "project_id": task_instance.project_id,
            },
        )


class VertexAIBatchPredictionJobLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI BatchPredictionJob link"""

    name = "Batch Prediction Job"
    key = "batch_prediction_job_conf"
    format_str = VERTEX_AI_BATCH_PREDICTION_JOB_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
        batch_prediction_job_id: str,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIBatchPredictionJobLink.key,
            value={
                "batch_prediction_job_id": batch_prediction_job_id,
                "region": task_instance.region,
                "project_id": task_instance.project_id,
            },
        )


class VertexAIBatchPredictionJobListLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI BatchPredictionJobList link"""

    name = "Batch Prediction Job List"
    key = "batch_prediction_jobs_conf"
    format_str = VERTEX_AI_BATCH_PREDICTION_JOB_LIST_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIBatchPredictionJobListLink.key,
            value={
                "project_id": task_instance.project_id,
            },
        )


class VertexAIEndpointLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI Endpoint link"""

    name = "Endpoint"
    key = "endpoint_conf"
    format_str = VERTEX_AI_ENDPOINT_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
        endpoint_id: str,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIEndpointLink.key,
            value={
                "endpoint_id": endpoint_id,
                "region": task_instance.region,
                "project_id": task_instance.project_id,
            },
        )


class VertexAIEndpointListLink(BaseGoogleLink):
    """Helper class for constructing Vertex AI EndpointList link"""

    name = "Endpoint List"
    key = "endpoints_conf"
    format_str = VERTEX_AI_ENDPOINT_LIST_LINK

    @staticmethod
    def persist(
        context: "Context",
        task_instance,
    ):
        task_instance.xcom_push(
            context=context,
            key=VertexAIEndpointListLink.key,
            value={
                "project_id": task_instance.project_id,
            },
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that displays interactions with Google Cloud Build.

This DAG relies on the following OS environment variables:

* GCP_PROJECT_ID - Google Cloud Project to use for the Cloud Function.
* GCP_CLOUD_BUILD_ARCHIVE_URL - Path to the zipped source in Google Cloud Storage.
    This object must be a gzipped archive file (.tar.gz) containing source to build.
* GCP_CLOUD_BUILD_REPOSITORY_NAME - Name of the Cloud Source Repository.

"""

import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict

import yaml
from future.backports.urllib.parse import urlparse

from airflow import models
from airflow.models.baseoperator import chain
from airflow.operators.bash import BashOperator
from airflow.providers.google.cloud.operators.cloud_build import (
    CloudBuildCancelBuildOperator,
    CloudBuildCreateBuildOperator,
    CloudBuildCreateBuildTriggerOperator,
    CloudBuildDeleteBuildTriggerOperator,
    CloudBuildGetBuildOperator,
    CloudBuildGetBuildTriggerOperator,
    CloudBuildListBuildsOperator,
    CloudBuildListBuildTriggersOperator,
    CloudBuildRetryBuildOperator,
    CloudBuildRunBuildTriggerOperator,
    CloudBuildUpdateBuildTriggerOperator,
)

START_DATE = datetime(2021, 1, 1)

GCP_PROJECT_ID = os.environ.get("GCP_PROJECT_ID", "aitflow-test-project")

GCP_SOURCE_ARCHIVE_URL = os.environ.get("GCP_CLOUD_BUILD_ARCHIVE_URL", "gs://airflow-test-bucket/file.tar.gz")
GCP_SOURCE_REPOSITORY_NAME = os.environ.get("GCP_CLOUD_BUILD_REPOSITORY_NAME", "airflow-test-repo")

GCP_SOURCE_ARCHIVE_URL_PARTS = urlparse(GCP_SOURCE_ARCHIVE_URL)
GCP_SOURCE_BUCKET_NAME = GCP_SOURCE_ARCHIVE_URL_PARTS.netloc

CURRENT_FOLDER = Path(__file__).parent

# [START howto_operator_gcp_create_build_trigger_body]
create_build_trigger_body = {
    "name": "test-cloud-build-trigger",
    "trigger_template": {
        "project_id": GCP_PROJECT_ID,
        "repo_name": GCP_SOURCE_REPOSITORY_NAME,
        "branch_name": "master",
    },
    "filename": "cloudbuild.yaml",
}
# [END howto_operator_gcp_create_build_trigger_body]

update_build_trigger_body = {
    "name": "test-cloud-build-trigger",
    "trigger_template": {
        "project_id": GCP_PROJECT_ID,
        "repo_name": GCP_SOURCE_REPOSITORY_NAME,
        "branch_name": "dev",
    },
    "filename": "cloudbuild.yaml",
}

# [START howto_operator_gcp_create_build_from_storage_body]
create_build_from_storage_body = {
    "source": {"storage_source": GCP_SOURCE_ARCHIVE_URL},
    "steps": [
        {
            "name": "gcr.io/cloud-builders/docker",
            "args": ["build", "-t", f"gcr.io/$PROJECT_ID/{GCP_SOURCE_BUCKET_NAME}", "."],
        }
    ],
    "images": [f"gcr.io/$PROJECT_ID/{GCP_SOURCE_BUCKET_NAME}"],
}
# [END howto_operator_gcp_create_build_from_storage_body]

# [START howto_operator_create_build_from_repo_body]
create_build_from_repo_body: Dict[str, Any] = {
    "source": {"repo_source": {"repo_name": GCP_SOURCE_REPOSITORY_NAME, "branch_name": "main"}},
    "steps": [
        {
            "name": "gcr.io/cloud-builders/docker",
            "args": ["build", "-t", "gcr.io/$PROJECT_ID/$REPO_NAME", "."],
        }
    ],
    "images": ["gcr.io/$PROJECT_ID/$REPO_NAME"],
}
# [END howto_operator_create_build_from_repo_body]


with models.DAG(
    "example_gcp_cloud_build",
    schedule_interval='@once',
    start_date=START_DATE,
    catchup=False,
    tags=["example"],
) as build_dag:

    # [START howto_operator_create_build_from_storage]
    create_build_from_storage = CloudBuildCreateBuildOperator(
        task_id="create_build_from_storage", project_id=GCP_PROJECT_ID, build=create_build_from_storage_body
    )
    # [END howto_operator_create_build_from_storage]

    # [START howto_operator_create_build_from_storage_result]
    create_build_from_storage_result = BashOperator(
        bash_command=f"echo { create_build_from_storage.output['results'] }",
        task_id="create_build_from_storage_result",
    )
    # [END howto_operator_create_build_from_storage_result]

    # [START howto_operator_create_build_from_repo]
    create_build_from_repo = CloudBuildCreateBuildOperator(
        task_id="create_build_from_repo", project_id=GCP_PROJECT_ID, build=create_build_from_repo_body
    )
    # [END howto_operator_create_build_from_repo]

    # [START howto_operator_create_build_from_repo_result]
    create_build_from_repo_result = BashOperator(
        bash_command=f"echo { create_build_from_repo.output['results'] }",
        task_id="create_build_from_repo_result",
    )
    # [END howto_operator_create_build_from_repo_result]

    # [START howto_operator_list_builds]
    list_builds = CloudBuildListBuildsOperator(
        task_id="list_builds", project_id=GCP_PROJECT_ID, location="global"
    )
    # [END howto_operator_list_builds]

    # [START howto_operator_create_build_without_wait]
    create_build_without_wait = CloudBuildCreateBuildOperator(
        task_id="create_build_without_wait",
        project_id=GCP_PROJECT_ID,
        build=create_build_from_repo_body,
        wait=False,
    )
    # [END howto_operator_create_build_without_wait]

    # [START howto_operator_cancel_build]
    cancel_build = CloudBuildCancelBuildOperator(
        task_id="cancel_build",
        id_=create_build_without_wait.output['id'],
        project_id=GCP_PROJECT_ID,
    )
    # [END howto_operator_cancel_build]

    # [START howto_operator_retry_build]
    retry_build = CloudBuildRetryBuildOperator(
        task_id="retry_build",
        id_=cancel_build.output['id'],
        project_id=GCP_PROJECT_ID,
    )
    # [END howto_operator_retry_build]

    # [START howto_operator_get_build]
    get_build = CloudBuildGetBuildOperator(
        task_id="get_build",
        id_=retry_build.output['id'],
        project_id=GCP_PROJECT_ID,
    )
    # [END howto_operator_get_build]

    # [START howto_operator_gcp_create_build_from_yaml_body]
    create_build_from_file = CloudBuildCreateBuildOperator(
        task_id="create_build_from_file",
        project_id=GCP_PROJECT_ID,
        build=yaml.safe_load((Path(CURRENT_FOLDER) / 'example_cloud_build.yaml').read_text()),
        params={'name': 'Airflow'},
    )
    # [END howto_operator_gcp_create_build_from_yaml_body]

    create_build_from_storage >> create_build_from_storage_result
    create_build_from_storage_result >> list_builds
    create_build_from_repo >> create_build_from_repo_result
    create_build_from_repo_result >> list_builds
    list_builds >> create_build_without_wait >> cancel_build
    cancel_build >> retry_build >> get_build

with models.DAG(
    "example_gcp_cloud_build_trigger",
    schedule_interval='@once',
    start_date=START_DATE,
    catchup=False,
    tags=["example"],
) as build_trigger_dag:

    # [START howto_operator_create_build_trigger]
    create_build_trigger = CloudBuildCreateBuildTriggerOperator(
        task_id="create_build_trigger", project_id=GCP_PROJECT_ID, trigger=create_build_trigger_body
    )
    # [END howto_operator_create_build_trigger]

    # [START howto_operator_run_build_trigger]
    run_build_trigger = CloudBuildRunBuildTriggerOperator(
        task_id="run_build_trigger",
        project_id=GCP_PROJECT_ID,
        trigger_id=create_build_trigger.output['id'],
        source=create_build_from_repo_body['source']['repo_source'],
    )
    # [END howto_operator_run_build_trigger]

    # [START howto_operator_create_build_trigger]
    update_build_trigger = CloudBuildUpdateBuildTriggerOperator(
        task_id="update_build_trigger",
        project_id=GCP_PROJECT_ID,
        trigger_id=create_build_trigger.output['id'],
        trigger=update_build_trigger_body,
    )
    # [END howto_operator_create_build_trigger]

    # [START howto_operator_get_build_trigger]
    get_build_trigger = CloudBuildGetBuildTriggerOperator(
        task_id="get_build_trigger",
        project_id=GCP_PROJECT_ID,
        trigger_id=create_build_trigger.output['id'],
    )
    # [END howto_operator_get_build_trigger]

    # [START howto_operator_delete_build_trigger]
    delete_build_trigger = CloudBuildDeleteBuildTriggerOperator(
        task_id="delete_build_trigger",
        project_id=GCP_PROJECT_ID,
        trigger_id=create_build_trigger.output['id'],
    )
    # [END howto_operator_delete_build_trigger]

    # [START howto_operator_list_build_triggers]
    list_build_triggers = CloudBuildListBuildTriggersOperator(
        task_id="list_build_triggers", project_id=GCP_PROJECT_ID, location="global", page_size=5
    )
    # [END howto_operator_list_build_triggers]

    chain(
        create_build_trigger,
        run_build_trigger,
        update_build_trigger,
        get_build_trigger,
        delete_build_trigger,
        list_build_triggers,
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that shows how to use DataFusion.
"""
import os
from datetime import datetime

from airflow import models
from airflow.operators.bash import BashOperator
from airflow.providers.google.cloud.operators.datafusion import (
    CloudDataFusionCreateInstanceOperator,
    CloudDataFusionCreatePipelineOperator,
    CloudDataFusionDeleteInstanceOperator,
    CloudDataFusionDeletePipelineOperator,
    CloudDataFusionGetInstanceOperator,
    CloudDataFusionListPipelinesOperator,
    CloudDataFusionRestartInstanceOperator,
    CloudDataFusionStartPipelineOperator,
    CloudDataFusionStopPipelineOperator,
    CloudDataFusionUpdateInstanceOperator,
)
from airflow.providers.google.cloud.sensors.datafusion import CloudDataFusionPipelineStateSensor

# [START howto_data_fusion_env_variables]
SERVICE_ACCOUNT = os.environ.get("GCP_DATAFUSION_SERVICE_ACCOUNT")
LOCATION = "europe-north1"
INSTANCE_NAME = "airflow-test-instance"
INSTANCE = {
    "type": "BASIC",
    "displayName": INSTANCE_NAME,
    "dataprocServiceAccount": SERVICE_ACCOUNT,
}

BUCKET_1 = os.environ.get("GCP_DATAFUSION_BUCKET_1", "test-datafusion-bucket-1")
BUCKET_2 = os.environ.get("GCP_DATAFUSION_BUCKET_2", "test-datafusion-bucket-2")

BUCKET_1_URI = f"gs://{BUCKET_1}"
BUCKET_2_URI = f"gs://{BUCKET_2}"

PIPELINE_NAME = os.environ.get("GCP_DATAFUSION_PIPELINE_NAME", "airflow_test")
PIPELINE = {
    "artifact": {
        "name": "cdap-data-pipeline",
        "version": "6.5.1",
        "scope": "SYSTEM",
        "label": "Data Pipeline - System Test",
    },
    "description": "Data Pipeline Application",
    "name": "test-pipe",
    "config": {
        "resources": {"memoryMB": 2048, "virtualCores": 1},
        "driverResources": {"memoryMB": 2048, "virtualCores": 1},
        "connections": [{"from": "GCS", "to": "GCS2"}],
        "comments": [],
        "postActions": [],
        "properties": {},
        "processTimingEnabled": "true",
        "stageLoggingEnabled": "false",
        "stages": [
            {
                "name": "GCS",
                "plugin": {
                    "name": "GCSFile",
                    "type": "batchsource",
                    "label": "GCS",
                    "artifact": {"name": "google-cloud", "version": "0.18.1", "scope": "SYSTEM"},
                    "properties": {
                        "project": "auto-detect",
                        "format": "text",
                        "skipHeader": "false",
                        "serviceFilePath": "auto-detect",
                        "filenameOnly": "false",
                        "recursive": "false",
                        "encrypted": "false",
                        "schema": "{\"type\":\"record\",\"name\":\"textfile\",\"fields\":[{\"name\"\
                            :\"offset\",\"type\":\"long\"},{\"name\":\"body\",\"type\":\"string\"}]}",
                        "path": BUCKET_1_URI,
                        "referenceName": "foo_bucket",
                        "useConnection": "false",
                        "serviceAccountType": "filePath",
                        "sampleSize": "1000",
                        "fileEncoding": "UTF-8",
                    },
                },
                "outputSchema": "{\"type\":\"record\",\"name\":\"textfile\",\"fields\"\
                    :[{\"name\":\"offset\",\"type\":\"long\"},{\"name\":\"body\",\"type\":\"string\"}]}",
                "id": "GCS",
            },
            {
                "name": "GCS2",
                "plugin": {
                    "name": "GCS",
                    "type": "batchsink",
                    "label": "GCS2",
                    "artifact": {"name": "google-cloud", "version": "0.18.1", "scope": "SYSTEM"},
                    "properties": {
                        "project": "auto-detect",
                        "suffix": "yyyy-MM-dd-HH-mm",
                        "format": "json",
                        "serviceFilePath": "auto-detect",
                        "location": "us",
                        "schema": "{\"type\":\"record\",\"name\":\"textfile\",\"fields\":[{\"name\"\
                            :\"offset\",\"type\":\"long\"},{\"name\":\"body\",\"type\":\"string\"}]}",
                        "referenceName": "bar",
                        "path": BUCKET_2_URI,
                        "serviceAccountType": "filePath",
                        "contentType": "application/octet-stream",
                    },
                },
                "outputSchema": "{\"type\":\"record\",\"name\":\"textfile\",\"fields\"\
                    :[{\"name\":\"offset\",\"type\":\"long\"},{\"name\":\"body\",\"type\":\"string\"}]}",
                "inputSchema": [
                    {
                        "name": "GCS",
                        "schema": "{\"type\":\"record\",\"name\":\"textfile\",\"fields\":[{\"name\"\
                            :\"offset\",\"type\":\"long\"},{\"name\":\"body\",\"type\":\"string\"}]}",
                    }
                ],
                "id": "GCS2",
            },
        ],
        "schedule": "0 * * * *",
        "engine": "spark",
        "numOfRecordsPreview": 100,
        "description": "Data Pipeline Application",
        "maxConcurrentRuns": 1,
    },
}
# [END howto_data_fusion_env_variables]


with models.DAG(
    "example_data_fusion",
    schedule_interval='@once',  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    # [START howto_cloud_data_fusion_create_instance_operator]
    create_instance = CloudDataFusionCreateInstanceOperator(
        location=LOCATION,
        instance_name=INSTANCE_NAME,
        instance=INSTANCE,
        task_id="create_instance",
    )
    # [END howto_cloud_data_fusion_create_instance_operator]

    # [START howto_cloud_data_fusion_get_instance_operator]
    get_instance = CloudDataFusionGetInstanceOperator(
        location=LOCATION, instance_name=INSTANCE_NAME, task_id="get_instance"
    )
    # [END howto_cloud_data_fusion_get_instance_operator]

    # [START howto_cloud_data_fusion_restart_instance_operator]
    restart_instance = CloudDataFusionRestartInstanceOperator(
        location=LOCATION, instance_name=INSTANCE_NAME, task_id="restart_instance"
    )
    # [END howto_cloud_data_fusion_restart_instance_operator]

    # [START howto_cloud_data_fusion_update_instance_operator]
    update_instance = CloudDataFusionUpdateInstanceOperator(
        location=LOCATION,
        instance_name=INSTANCE_NAME,
        instance=INSTANCE,
        update_mask="",
        task_id="update_instance",
    )
    # [END howto_cloud_data_fusion_update_instance_operator]

    # [START howto_cloud_data_fusion_create_pipeline]
    create_pipeline = CloudDataFusionCreatePipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        pipeline=PIPELINE,
        instance_name=INSTANCE_NAME,
        task_id="create_pipeline",
    )
    # [END howto_cloud_data_fusion_create_pipeline]

    # [START howto_cloud_data_fusion_list_pipelines]
    list_pipelines = CloudDataFusionListPipelinesOperator(
        location=LOCATION, instance_name=INSTANCE_NAME, task_id="list_pipelines"
    )
    # [END howto_cloud_data_fusion_list_pipelines]

    # [START howto_cloud_data_fusion_start_pipeline]
    start_pipeline = CloudDataFusionStartPipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        task_id="start_pipeline",
    )
    # [END howto_cloud_data_fusion_start_pipeline]

    # [START howto_cloud_data_fusion_start_pipeline_async]
    start_pipeline_async = CloudDataFusionStartPipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        asynchronous=True,
        task_id="start_pipeline_async",
    )

    # [END howto_cloud_data_fusion_start_pipeline_async]

    # [START howto_cloud_data_fusion_start_pipeline_sensor]
    start_pipeline_sensor = CloudDataFusionPipelineStateSensor(
        task_id="pipeline_state_sensor",
        pipeline_name=PIPELINE_NAME,
        pipeline_id=start_pipeline_async.output,
        expected_statuses=["COMPLETED"],
        failure_statuses=["FAILED"],
        instance_name=INSTANCE_NAME,
        location=LOCATION,
    )
    # [END howto_cloud_data_fusion_start_pipeline_sensor]

    # [START howto_cloud_data_fusion_stop_pipeline]
    stop_pipeline = CloudDataFusionStopPipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        task_id="stop_pipeline",
    )
    # [END howto_cloud_data_fusion_stop_pipeline]

    # [START howto_cloud_data_fusion_delete_pipeline]
    delete_pipeline = CloudDataFusionDeletePipelineOperator(
        location=LOCATION,
        pipeline_name=PIPELINE_NAME,
        instance_name=INSTANCE_NAME,
        task_id="delete_pipeline",
    )
    # [END howto_cloud_data_fusion_delete_pipeline]

    # [START howto_cloud_data_fusion_delete_instance_operator]
    delete_instance = CloudDataFusionDeleteInstanceOperator(
        location=LOCATION, instance_name=INSTANCE_NAME, task_id="delete_instance"
    )
    # [END howto_cloud_data_fusion_delete_instance_operator]

    # Add sleep before creating pipeline
    sleep = BashOperator(task_id="sleep", bash_command="sleep 60")

    create_instance >> get_instance >> restart_instance >> update_instance >> sleep
    (
        sleep
        >> create_pipeline
        >> list_pipelines
        >> start_pipeline_async
        >> start_pipeline_sensor
        >> start_pipeline
        >> stop_pipeline
        >> delete_pipeline
    )
    delete_pipeline >> delete_instance

if __name__ == "__main__":
    dag.clear()
    dag.run()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import asyncio
import logging
from typing import Any, Dict, Optional, Sequence, Tuple, Union

from airflow import AirflowException
from airflow.providers.google.cloud.hooks.cloud_composer import CloudComposerHook

try:
    from airflow.triggers.base import BaseTrigger, TriggerEvent
except ImportError:
    logging.getLogger(__name__).warning(
        'Deferrable Operators only work starting Airflow 2.2',
        exc_info=True,
    )
    BaseTrigger = object  # type: ignore
    TriggerEvent = None  # type: ignore


class CloudComposerExecutionTrigger(BaseTrigger):
    """The trigger handles the async communication with the Google Cloud Composer"""

    def __init__(
        self,
        project_id: str,
        region: str,
        operation_name: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        delegate_to: Optional[str] = None,
        pooling_period_seconds: int = 30,
    ):
        super().__init__()
        self.project_id = project_id
        self.region = region
        self.operation_name = operation_name

        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.delegate_to = delegate_to

        self.pooling_period_seconds = pooling_period_seconds

        self.gcp_hook = CloudComposerHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            delegate_to=self.delegate_to,
        )

    def serialize(self) -> Tuple[str, Dict[str, Any]]:
        return (
            'airflow.providers.google.cloud.triggers.cloud_composer.CloudComposerExecutionTrigger',
            {
                "project_id": self.project_id,
                "region": self.region,
                "operation_name": self.operation_name,
                "gcp_conn_id": self.gcp_conn_id,
                "impersonation_chain": self.impersonation_chain,
                "delegate_to": self.delegate_to,
                "pooling_period_seconds": self.pooling_period_seconds,
            },
        )

    async def run(self):
        while True:
            operation = self.gcp_hook.get_operation(operation_name=self.operation_name)
            if operation.done:
                break
            elif operation.error.message:
                raise AirflowException(f"Cloud Composer Environment error: {operation.error.message}")
            await asyncio.sleep(self.pooling_period_seconds)
        yield TriggerEvent(
            {
                "operation_name": operation.name,
                "operation_done": operation.done,
            }
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Video Intelligence Hook."""
from typing import Dict, List, Optional, Sequence, Tuple, Union

from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.api_core.operation import Operation
from google.api_core.retry import Retry
from google.cloud.videointelligence_v1 import VideoIntelligenceServiceClient
from google.cloud.videointelligence_v1.types import VideoContext

from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook


class CloudVideoIntelligenceHook(GoogleBaseHook):
    """
    Hook for Google Cloud Video Intelligence APIs.

    All the methods in the hook where project_id is used must be called with
    keyword arguments rather than positional.

    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param delegate_to: The account to impersonate using domain-wide delegation of authority,
        if any. For this to work, the service account making the request must have
        domain-wide delegation enabled.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        delegate_to: Optional[str] = None,
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            delegate_to=delegate_to,
            impersonation_chain=impersonation_chain,
        )
        self._conn = None

    def get_conn(self) -> VideoIntelligenceServiceClient:
        """
        Returns Gcp Video Intelligence Service client

        :rtype: google.cloud.videointelligence_v1.VideoIntelligenceServiceClient
        """
        if not self._conn:
            self._conn = VideoIntelligenceServiceClient(
                credentials=self._get_credentials(), client_info=CLIENT_INFO
            )
        return self._conn

    @GoogleBaseHook.quota_retry()
    def annotate_video(
        self,
        input_uri: Optional[str] = None,
        input_content: Optional[bytes] = None,
        features: Optional[List[VideoIntelligenceServiceClient.enums.Feature]] = None,
        video_context: Union[Dict, VideoContext] = None,
        output_uri: Optional[str] = None,
        location: Optional[str] = None,
        retry: Union[Retry, _MethodDefault] = DEFAULT,
        timeout: Optional[float] = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> Operation:
        """
        Performs video annotation.

        :param input_uri: Input video location. Currently, only Google Cloud Storage URIs are supported,
            which must be specified in the following format: ``gs://bucket-id/object-id``.
        :param input_content: The video data bytes.
            If unset, the input video(s) should be specified via ``input_uri``.
            If set, ``input_uri`` should be unset.
        :param features: Requested video annotation features.
        :param output_uri: Optional, location where the output (in JSON format) should be stored. Currently,
            only Google Cloud Storage URIs are supported, which must be specified in the following format:
            ``gs://bucket-id/object-id``.
        :param video_context: Optional, Additional video context and/or feature-specific parameters.
        :param location: Optional, cloud region where annotation should take place. Supported cloud regions:
            us-east1, us-west1, europe-west1, asia-east1.
            If no region is specified, a region will be determined based on video file location.
        :param retry: Retry object used to determine when/if to retry requests.
            If None is specified, requests will not be retried.
        :param timeout: Optional, The amount of time, in seconds, to wait for the request to complete.
            Note that if retry is specified, the timeout applies to each individual attempt.
        :param metadata: Optional, Additional metadata that is provided to the method.
        """
        client = self.get_conn()
        return client.annotate_video(
            input_uri=input_uri,
            input_content=input_content,
            features=features,
            video_context=video_context,
            output_uri=output_uri,
            location_id=location,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Bigquery sensor."""
from typing import TYPE_CHECKING, Optional, Sequence, Union

from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from airflow.sensors.base import BaseSensorOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class BigQueryTableExistenceSensor(BaseSensorOperator):
    """
    Checks for the existence of a table in Google Bigquery.

    :param project_id: The Google cloud project in which to look for the table.
        The connection supplied to the hook must provide
        access to the specified project.
    :param dataset_id: The name of the dataset in which to look for the table.
        storage bucket.
    :param table_id: The name of the table to check the existence of.
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param delegate_to: The account to impersonate using domain-wide delegation of authority,
        if any. For this to work, the service account making the request must have
        domain-wide delegation enabled.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        'project_id',
        'dataset_id',
        'table_id',
        'impersonation_chain',
    )
    ui_color = '#f0eee4'

    def __init__(
        self,
        *,
        project_id: str,
        dataset_id: str,
        table_id: str,
        gcp_conn_id: str = 'google_cloud_default',
        delegate_to: Optional[str] = None,
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:

        super().__init__(**kwargs)

        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to
        self.impersonation_chain = impersonation_chain

    def poke(self, context: 'Context') -> bool:
        table_uri = f'{self.project_id}:{self.dataset_id}.{self.table_id}'
        self.log.info('Sensor checks existence of table: %s', table_uri)
        hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id,
            delegate_to=self.delegate_to,
            impersonation_chain=self.impersonation_chain,
        )
        return hook.table_exists(
            project_id=self.project_id, dataset_id=self.dataset_id, table_id=self.table_id
        )


class BigQueryTablePartitionExistenceSensor(BaseSensorOperator):
    """
    Checks for the existence of a partition within a table in Google Bigquery.

    :param project_id: The Google cloud project in which to look for the table.
        The connection supplied to the hook must provide
        access to the specified project.
    :param dataset_id: The name of the dataset in which to look for the table.
        storage bucket.
    :param table_id: The name of the table to check the existence of.
    :param partition_id: The name of the partition to check the existence of.
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param delegate_to: The account to impersonate, if any.
        For this to work, the service account making the request must
        have domain-wide delegation enabled.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        'project_id',
        'dataset_id',
        'table_id',
        'partition_id',
        'impersonation_chain',
    )
    ui_color = '#f0eee4'

    def __init__(
        self,
        *,
        project_id: str,
        dataset_id: str,
        table_id: str,
        partition_id: str,
        gcp_conn_id: str = 'google_cloud_default',
        delegate_to: Optional[str] = None,
        impersonation_chain: Optional[Union[str, Sequence[str]]] = None,
        **kwargs,
    ) -> None:

        super().__init__(**kwargs)

        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.partition_id = partition_id
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to
        self.impersonation_chain = impersonation_chain

    def poke(self, context: 'Context') -> bool:
        table_uri = f'{self.project_id}:{self.dataset_id}.{self.table_id}'
        self.log.info('Sensor checks existence of partition: "%s" in table: %s', self.partition_id, table_uri)
        hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id,
            delegate_to=self.delegate_to,
            impersonation_chain=self.impersonation_chain,
        )
        return hook.table_partition_exists(
            project_id=self.project_id,
            dataset_id=self.dataset_id,
            table_id=self.table_id,
            partition_id=self.partition_id,
        )
